#!/usr/bin/python
"""
Expose a CloudFileFS object over SFTP using paramiko

Copyright (C) 2011-2013 by Memset Ltd. http://www.memset.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import logging

import os
import errno
import shlex
from SocketServer import StreamRequestHandler, ForkingTCPServer

import paramiko
from Crypto import Random

from ftpcloudfs.fs import ObjectStorageFS, ObjectStorageFD
from ftpcloudfs.utils import smart_str
from sftpcloudfs.scp import SCPHandler

from functools import wraps
from posixpath import basename

def return_sftp_errors(func):
    """
    Decorator to catch EnvironmentError~s and return SFTP error codes instead.

    Other exceptions are logged and processed as EIO errors.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        log = paramiko.util.get_logger("paramiko")
        name = getattr(func, "func_name", "unknown")
        try:
            log.debug("%s(%r,%r): enter" % (name, args, kwargs))
            rc = func(*args, **kwargs)
        except BaseException, e:
            obj = args[0]
            params = args[1:] if len(args) > 1 else ()
            msg = "%s%r from %r: %s" % (name, params, obj.client_address, e)
            if isinstance(e, EnvironmentError):
                log.info(msg)
                error = e.errno
            else:
                log.exception("unexpected error: %s" % msg)
                error = errno.EIO
            rc = paramiko.SFTPServer.convert_errno(error)
        log.debug("%s: returns %r" % (name, rc))
        return rc
    return wrapper


class SFTPServerInterface(paramiko.SFTPServerInterface):
    """
    SFTPServerInterface implementation that exposes a ObjectStorageFS object.
    """

    def __init__(self, server, fs, *args, **kwargs):
        self.fs = fs
        self.client_address = server.client_address
        self.log = paramiko.util.get_logger("paramiko")
        self.log.debug("%s: start filesystem interface" % self.__class__.__name__)
        super(SFTPServerInterface,self).__init__(server, *args, **kwargs)

    @return_sftp_errors
    def open(self, path, flags, attr):
        return SFTPHandle(self, path, flags)

    @return_sftp_errors
    def list_folder(self, path):
        return [ paramiko.SFTPAttributes.from_stat(stat, smart_str(leaf))
                 for leaf, stat in self.fs.listdir_with_stat(path) ]

    @return_sftp_errors
    def stat(self, path):
        stat = self.fs.stat(path)
        filename = basename(path)
        return paramiko.SFTPAttributes.from_stat(stat, filename)

    def lstat(self, path):
        return self.stat(path)

    @return_sftp_errors
    def remove(self, path):
        self.fs.remove(path)
        return paramiko.SFTP_OK

    @return_sftp_errors
    def rename(self, oldpath, newpath):
        self.fs.rename(oldpath, newpath)
        return paramiko.SFTP_OK

    @return_sftp_errors
    def mkdir(self, path, attr):
        self.fs.mkdir(path)
        return paramiko.SFTP_OK

    @return_sftp_errors
    def rmdir(self, path):
        self.fs.rmdir(path)
        return paramiko.SFTP_OK

    def canonicalize(self, path):
        return smart_str(self.fs.abspath(self.fs.normpath(path)))

    @return_sftp_errors
    def chattr(self, path, attr):
        return paramiko.SFTP_OP_UNSUPPORTED

    def readlink(self, path):
        return paramiko.SFTP_OP_UNSUPPORTED

    def symlink(self, path):
        return paramiko.SFTP_OP_UNSUPPORTED


class SFTPHandle(paramiko.SFTPHandle):
    """
    Expose a ObjectStorageFD object to SFTP.
    """

    def __init__(self, owner, path, flags):
        super(SFTPHandle, self).__init__(flags)
        self.log = paramiko.util.get_logger("paramiko")
        self.owner = owner
        self.path = path
        self.log.debug("SFTPHandle(path=%r, flags=%r)" % (path, flags))
        open_mode = flags & (os.O_RDONLY | os.O_WRONLY | os.O_RDWR)
        if open_mode == os.O_RDONLY:
            mode = "r"
        elif open_mode == os.O_WRONLY:
            mode = "w"
        elif open_mode == os.O_RDWR:
            mode = "rw"
        else:
            self.log.error("Bad open mode %r" % flags)
            return paramiko.SFTP_OP_UNSUPPORTED
        if flags & os.O_APPEND:
            mode += "+"

        # we need the file size for r & rw mode; this needs to be performed
        # BEFORE open so the cache gets invalidated in write operations
        try:
            self._size = owner.fs.stat(path).st_size
        except EnvironmentError:
            self._size = 0

        # FIXME ignores os.O_CREAT, os.O_TRUNC, os.O_EXCL
        self._file = owner.fs.open(path, mode)
        self._tell = 0

    @property
    def client_address(self):
        return self.owner.client_address

    @return_sftp_errors
    def close(self):
        self._file.close()
        return paramiko.SFTP_OK

    @return_sftp_errors
    def read(self, offset, length):
        if offset != self._tell:
            # this is not an "invalid offset" error
            if offset > self._size:
                return paramiko.SFTP_EOF
            self._file.seek(offset)
            self._tell = offset
        data = self._file.read(length)
        self._tell += len(data)
        return data

    @return_sftp_errors
    def write(self, offset, data):
        if offset != self._tell:
            return paramiko.SFTP_OP_UNSUPPORTED
            # FIXME self._file.seek(offset)
        self._file.write(data)
        self._tell += len(data)
        # update the file size
        if self._tell > self._size:
            self._size = self._tell
        return paramiko.SFTP_OK

    def stat(self):
        return self.owner.stat(self.path)

    def chattr(self,attr):
        return paramiko.SFTP_OP_UNSUPPORTED


class ObjectStorageSFTPRequestHandler(StreamRequestHandler):
    """
    SocketServer RequestHandler subclass for ObjectStorageSFTPServer.

    This RequestHandler subclass creates a paramiko Transport, sets up the
    sftp subsystem, and hands off to the transport's own request handling
    thread.  Note that paramiko.Transport uses a separate thread by default,
    so there is no need to use ThreadingMixin.

    A TERM signal may be processed with a delay up to 10 seconds.
    """

    timeout = 60
    auth_timeout = 60

    def handle(self):
        Random.atfork()
        paramiko.util.get_logger("paramiko.transport").setLevel(logging.CRITICAL)
        self.log = paramiko.util.get_logger("paramiko")
        self.log.debug("%s: start transport" % self.__class__.__name__)
        self.server.client_address = self.client_address
        t = paramiko.Transport(self.request)
        t.add_server_key(self.server.host_key)
        t.set_subsystem_handler("sftp", paramiko.SFTPServer, SFTPServerInterface, self.server.fs)
        try:
            t.start_server(server=self.server)
        except paramiko.SSHException, e:
            self.log.warning("Disconnecting: %s" % e)
            t.close()
            return
        chan = t.accept(self.auth_timeout)
        if chan is None:
            self.log.warning("Channel is None, closing")
            t.close()
            return

        while t.isAlive():
            t.join(timeout=10)

class ObjectStorageSFTPServer(ForkingTCPServer, paramiko.ServerInterface):
    """
    Expose a ObjectStorageFS object over SFTP.
    """
    allow_reuse_address = True

    def __init__(self, address, host_key=None, authurl=None, max_children=20, keystone=None, no_scp=False, split_size=0):
        self.log = paramiko.util.get_logger("paramiko")
        self.log.debug("%s: start server" % self.__class__.__name__)
        self.fs = ObjectStorageFS(None, None, authurl=authurl, keystone=keystone) # unauthorized
        self.host_key = host_key
        self.max_children = max_children
        self.no_scp = no_scp
        ForkingTCPServer.__init__(self, address, ObjectStorageSFTPRequestHandler)
        ObjectStorageFD.split_size = split_size

    def check_channel_request(self, kind, chanid):
        if kind == 'session':
            return paramiko.OPEN_SUCCEEDED
        self.log.warning("Channel request denied from %s, kind=%s" \
                         % (self.client_address, kind))
        # all the check_channel_*_request return False by default but
        # sftp subsystem because of the set_subsystem_handler call in
        # the ObjectStorageSFTPRequestHandler
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_channel_exec_request(self, channel, command):
        """Determine if a shell command will be executed for the client."""

        # Parse the command
        if ' -- ' in command:
            # scp will use -- to delimit the begining of the unscaped filename
            # so translate it to something that shelex can manage
            command = command.replace(' -- ', ' "') + '"'
        command = shlex.split(command)
        self.log.debug('check_channel_exec_request %r' % command)

        try:
            if command[0] == 'scp':
                if self.no_scp:
                    self.log.info("scp exec request denied from=%s (scp is disabled)" % (self.client_address,))
                    return False
                self.log.info('invoking %r from=%s' % (command, self.client_address))
                # handle the command execution
                SCPHandler(command[1:], channel, self.fs, self.log).start()
                return True
        except:
            self.log.exception("command %r failed from=%s" % (command, self.client_address))
            return False

        return False

    def check_auth_none(self, username):
        """Check whether the user can proceed without authentication."""
        return paramiko.AUTH_FAILED

    def check_auth_publickey(self, username, key):
        """Check whether the given public key is valid for authentication."""
        return paramiko.AUTH_FAILED

    def check_auth_password(self, username, password):
        """Check whether the given password is valid for authentication."""
        self.log.info("Auth request (type=password), username=%s, from=%s" \
                      % (username, self.client_address))
        try:
            if not password:
                raise EnvironmentError("no password provided")
            self.fs.authenticate(username, password)
        except EnvironmentError, e:
            self.log.warning("%s: Failed to authenticate: %s" % (self.client_address, e))
            self.log.error("Authentication failure for %s from %s port %s" % (username,
                           self.client_address[0], self.client_address[1]))
            return paramiko.AUTH_FAILED
        self.fs.conn.real_ip = self.client_address[0]
        self.log.info("%s authenticated from %s" % (username, self.client_address))
        return paramiko.AUTH_SUCCESSFUL

    def get_allowed_auths(self,username):
        """Return string containing a comma separated list of allowed auth modes.

        The available modes are  "node", "password" and "publickey".
        """
        return "password"




#!/usr/bin/python
import unittest
import os
import sys
from time import time
from swiftclient import client
import paramiko
import stat

hostname = "127.0.0.1"
port = 8022

# setup logging
#paramiko.util.log_to_file('test_sftpd.log')

class SftpcloudfsTest(unittest.TestCase):
    ''' FTP Cloud FS main test '''

    def setUp(self):
        if not all(['OS_API_KEY' in os.environ,
                    'OS_API_USER' in os.environ,
                    'OS_AUTH_URL' in os.environ,
                    ]):
            print "env OS_API_USER/OS_API_KEY/OS_AUTH_URL not found."
            sys.exit(1)

        self.username = os.environ['OS_API_USER']
        self.api_key = os.environ['OS_API_KEY']
        self.auth_url = os.environ.get('OS_AUTH_URL')

        self.transport = paramiko.Transport((hostname, port))
        self.transport.connect(username=self.username, password=self.api_key) #, hostkey=hostkey)
        self.sftp = paramiko.SFTPClient.from_transport(self.transport)
        self.container = "sftpcloudfs_testing"
        self.sftp.mkdir("/%s" % self.container)
        self.sftp.chdir("/%s" % self.container)
        self.conn = client.Connection(user=self.username, key=self.api_key, authurl=self.auth_url)

    def create_file(self, path, contents):
        '''Create path with contents'''
        fd = self.sftp.open(path, "w")
        try:
            fd.write(contents)
        finally:
            fd.close()

    def read_file(self, path):
        '''Read contents of path'''
        fd = self.sftp.open(path, "r")
        try:
            return fd.read()
        finally:
            fd.close()

    def test_setup_and_teardown(self):
        pass

    def test_mkdir_chdir_rmdir(self):
        ''' mkdir/chdir/rmdir directory '''
        directory = "/foobarrandom"
        self.sftp.mkdir(directory)
        self.sftp.chdir(directory)
        self.assertEqual(self.sftp.getcwd(), directory)
        self.sftp.chdir("..")
        self.sftp.rmdir(directory)

    def test_mkdir_chdir_mkdir_rmdir_subdir(self):
        ''' mkdir/chdir/rmdir sub directory '''
        directory = "/foobarrandom"
        self.sftp.mkdir(directory)
        self.sftp.chdir(directory)
        self.assertEqual(self.sftp.getcwd(), directory)
        subdirectory = "potato"
        subdirpath = directory + "/" + subdirectory
        self.sftp.mkdir(subdirectory)
        # Can't delete a directory with stuff in
        self.assertRaises(EnvironmentError, self.sftp.rmdir, directory)
        self.sftp.chdir(subdirectory)
        self.assertEqual(self.sftp.getcwd(), subdirpath)
        self.sftp.chdir("..")
        self.assertEqual(self.sftp.getcwd(), directory)
        self.sftp.rmdir(subdirectory)
        self.sftp.chdir("..")
        self.sftp.rmdir(directory)

    def test_write_open_delete(self):
        ''' write/open/delete file '''
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.assertEquals(self.sftp.stat("testfile.txt").st_size, len(content_string))
        retrieved = self.read_file("testfile.txt")
        self.assertEqual(retrieved, content_string)
        self.sftp.remove("testfile.txt")

    def test_write_open_delete_subdir(self):
        ''' write/open/delete file in a subdirectory'''
        self.sftp.mkdir("potato")
        self.sftp.chdir("potato")
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.assertEquals(self.sftp.stat("testfile.txt").st_size, len(content_string))
        retrieved = self.read_file("/%s/potato/testfile.txt" % self.container)
        self.assertEqual(retrieved, content_string)
        self.sftp.remove("testfile.txt")
        self.sftp.chdir("..")
        self.sftp.rmdir("potato")

    def test_write_to_slash(self):
        ''' write to slash should not be permitted '''
        self.sftp.chdir("/")
        content_string = "Hello Moto"
        self.assertRaises(EnvironmentError, self.create_file, "testfile.txt", content_string)

    def test_chdir_to_a_file(self):
        ''' chdir to a file '''
        self.create_file("testfile.txt", "Hello Moto")
        self.assertRaises(paramiko.SFTPError, self.sftp.chdir, "/%s/testfile.txt" % self.container)
        self.sftp.remove("testfile.txt")

    def test_chdir_to_slash(self):
        ''' chdir to slash '''
        self.sftp.chdir("/")

    def test_chdir_to_nonexistent_container(self):
        ''' chdir to non existent container'''
        self.assertRaises(EnvironmentError, self.sftp.chdir, "/i_dont_exist")

    def test_chdir_to_nonexistent_directory(self):
        ''' chdir to nonexistend directory'''
        self.assertRaises(EnvironmentError, self.sftp.chdir, "i_dont_exist")
        self.assertRaises(EnvironmentError, self.sftp.chdir, "/%s/i_dont_exist" % self.container)

    def test_listdir_root(self):
        ''' list root directory '''
        self.sftp.chdir("/")
        ls = self.sftp.listdir()
        self.assertTrue(self.container in ls)
        self.assertTrue('potato' not in ls)
        self.sftp.mkdir("potato")
        ls = self.sftp.listdir()
        self.assertTrue(self.container in ls)
        self.assertTrue('potato' in ls)
        self.sftp.rmdir("potato")

    def test_listdir(self):
        ''' list directory '''
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.assertEqual(self.sftp.listdir(), ["testfile.txt"])
        self.sftp.remove("testfile.txt")

    def test_listdir_attr(self):
        ''' list directory '''
        content_string = "Hello Moto"
        start = time()
        self.create_file("testfile.txt", content_string)
        ls = self.sftp.listdir_attr()
        self.assertEqual(len(ls), 1)
        ls = ls[0]
        self.assertEqual(ls.filename, "testfile.txt")
        self.assertEqual(ls.st_size, 10)
        dt = abs(ls.st_mtime - start)
        self.assertTrue(dt < 5.0, "timestamp for file is out by %r seconds (%r - %r)" % (dt, ls.st_mtime, start))
        self.assertEquals(ls.st_mode, stat.S_IFREG + 0644)
        self.sftp.remove("testfile.txt")

    def test_listdir_subdir(self):
        ''' list a sub directory'''
        content_string = "Hello Moto"
        self.create_file("1.txt", content_string)
        self.create_file("2.txt", content_string)
        start = time()
        self.sftp.mkdir("potato")
        self.create_file("potato/3.txt", content_string)
        self.create_file("potato/4.txt", content_string)
        self.assertEqual(self.sftp.listdir(), ["1.txt", "2.txt", "potato"])
        self.sftp.chdir("potato")
        self.assertEqual(self.sftp.listdir(), ["3.txt", "4.txt"])
        self.sftp.remove("3.txt")
        self.sftp.remove("4.txt")
        self.assertEqual(self.sftp.listdir(), [])
        self.sftp.chdir("..")
        self.sftp.remove("1.txt")
        self.sftp.remove("2.txt")
        self.assertEqual(self.sftp.listdir(), ["potato"])
        ls = self.sftp.listdir_attr()
        self.assertEqual(len(ls), 1)
        ls = ls[0]
        self.assertEqual(ls.filename, "potato")
        self.assertEqual(ls.st_size, 0)
        dt = abs(ls.st_mtime - start)
        self.assertTrue(dt < 5.0, "timestamp for dir is out by %r seconds (%r - %r)" % (dt, ls.st_mtime, start))
        self.assertEquals(ls.st_mode, stat.S_IFDIR + 0755)
        self.sftp.rmdir("potato")
        self.assertEqual(self.sftp.listdir(), [])

    def test_rename_file(self):
        '''rename a file'''
        content_string = "Hello Moto" * 100
        self.create_file("testfile.txt", content_string)
        self.assertEquals(self.sftp.stat("testfile.txt").st_size, len(content_string))
        self.assertRaises(EnvironmentError, self.sftp.stat, "testfile2.txt")
        self.sftp.rename("testfile.txt", "testfile2.txt")
        self.assertEquals(self.sftp.stat("testfile2.txt").st_size, len(content_string))
        self.assertRaises(EnvironmentError, self.sftp.stat, "testfile.txt")
        self.sftp.remove("testfile2.txt")

    def test_rename_file_into_subdir1(self):
        '''rename a file into a subdirectory 1'''
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.sftp.mkdir("potato")
        self.assertEquals(self.sftp.stat("testfile.txt").st_size, len(content_string))
        self.assertRaises(EnvironmentError, self.sftp.stat, "potato/testfile3.txt")
        self.sftp.rename("testfile.txt", "potato/testfile3.txt")
        self.assertEquals(self.sftp.stat("potato/testfile3.txt").st_size, len(content_string))
        self.assertRaises(EnvironmentError, self.sftp.stat, "testfile.txt")
        self.sftp.remove("potato/testfile3.txt")
        self.sftp.rmdir("potato")

    def test_rename_file_into_subdir2(self):
        '''rename a file into a subdirectory without specifying dest leaf'''
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.sftp.mkdir("potato")
        self.assertEquals(self.sftp.stat("testfile.txt").st_size, len(content_string))
        self.assertRaises(EnvironmentError, self.sftp.stat, "potato/testfile.txt")
        self.sftp.rename("testfile.txt", "potato")
        self.assertEquals(self.sftp.stat("potato/testfile.txt").st_size, len(content_string))
        self.assertRaises(EnvironmentError, self.sftp.stat, "testfile.txt")
        self.sftp.remove("potato/testfile.txt")
        self.sftp.rmdir("potato")

    def test_rename_file_into_root(self):
        '''rename a file into a subdirectory without specifying dest leaf'''
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.assertRaises(EnvironmentError, self.sftp.rename, "testfile.txt", "/testfile.txt")
        self.sftp.remove("testfile.txt")

    def test_rename_directory_into_file(self):
        '''rename a directory into a file - shouldn't work'''
        content_string = "Hello Moto"
        self.create_file("testfile.txt", content_string)
        self.assertRaises(EnvironmentError, self.sftp.rename, "/%s" % self.container, "testfile.txt")
        self.sftp.remove("testfile.txt")

    def test_rename_directory_into_directory(self):
        '''rename a directory into a directory'''
        self.sftp.mkdir("potato")
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.sftp.rename("potato", "potato2")
        self.assertEquals(self.sftp.listdir("potato2"), [])
        self.sftp.rmdir("potato2")

    def test_rename_directory_into_existing_directory(self):
        '''rename a directory into an existing directory'''
        self.sftp.mkdir("potato")
        self.sftp.mkdir("potato2")
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.assertEquals(self.sftp.listdir("potato2"), [])
        self.sftp.rename("potato", "potato2")
        self.assertEquals(self.sftp.listdir("potato2"), ["potato"])
        self.assertEquals(self.sftp.listdir("potato2/potato"), [])
        self.sftp.rmdir("potato2/potato")
        self.sftp.rmdir("potato2")

    def test_rename_directory_into_self(self):
        '''rename a directory into itself'''
        self.sftp.mkdir("potato")
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.sftp.rename("potato", "/%s" % self.container)
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.sftp.rename("potato", "/%s/potato" % self.container)
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.sftp.rename("potato", "potato")
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.sftp.rename("/%s/potato" % self.container, ".")
        self.assertEquals(self.sftp.listdir("potato"), [])
        self.sftp.rmdir("potato")

    def test_rename_full_directory(self):
        '''rename a directory into a directory'''
        self.sftp.mkdir("potato")
        self.create_file("potato/something.txt", "p")
        try:
            self.assertEquals(self.sftp.listdir("potato"), ["something.txt"])
            self.assertRaises(EnvironmentError, self.sftp.rename, "potato", "potato2")
        finally:
            self.sftp.remove("potato/something.txt")
            self.sftp.rmdir("potato")

    def test_rename_container(self):
        '''rename an empty container'''
        self.sftp.mkdir("/potato")
        self.assertEquals(self.sftp.listdir("/potato"), [])
        self.assertRaises(EnvironmentError, self.sftp.listdir, "/potato2")
        self.sftp.rename("/potato", "/potato2")
        self.assertRaises(EnvironmentError, self.sftp.listdir, "/potato")
        self.assertEquals(self.sftp.listdir("/potato2"), [])
        self.sftp.rmdir("/potato2")

    def test_rename_full_container(self):
        '''rename a full container'''
        self.sftp.mkdir("/potato")
        self.create_file("/potato/test.txt", "onion")
        self.assertEquals(self.sftp.listdir("/potato"), ["test.txt"])
        self.assertRaises(EnvironmentError, self.sftp.rename, "/potato", "/potato2")
        self.sftp.remove("/potato/test.txt")
        self.sftp.rmdir("/potato")

    def test_unicode_file(self):
        '''Test unicode file creation'''
        file_name = u"Smiley\u263a.txt"
        self.create_file(file_name, "Hello Moto")
        self.assertEqual(self.sftp.listdir(), [file_name])
        self.sftp.remove(file_name)

    def test_unicode_directory(self):
        '''Test unicode directory creation'''
        dir_name = u"Smiley\u263aDir"
        self.sftp.mkdir(dir_name)
        self.assertEqual(self.sftp.listdir(), [dir_name])
        self.sftp.rmdir(dir_name)

    def test_mkdir_container_unicode(self):
        ''' mkdir/chdir/rmdir directory '''
        directory = u"/Smiley\u263aContainer"
        self.sftp.mkdir(directory)
        self.sftp.chdir(directory)
        # FIXME shouldn't have to utf-8 decode here?
        self.assertEqual(self.sftp.getcwd().decode("utf-8"), directory)
        self.sftp.chdir("..")
        self.sftp.rmdir(directory)

    def test_fakedir(self):
        '''Make some fake directories and test'''

        objs  = [ "test1.txt", "potato/test2.txt", "potato/sausage/test3.txt", "potato/sausage/test4.txt", ]
        for obj in objs:
            self.conn.put_object(self.container, obj, content_type="text/plain", contents="Hello Moto")

        self.assertEqual(self.sftp.listdir(), ["potato", "test1.txt"])
        self.assertEqual(self.sftp.listdir("potato"), ["sausage","test2.txt"])
        self.assertEqual(self.sftp.listdir("potato/sausage"), ["test3.txt", "test4.txt"])

        self.sftp.chdir("potato")

        self.assertEqual(self.sftp.listdir(), ["sausage","test2.txt"])
        self.assertEqual(self.sftp.listdir("sausage"), ["test3.txt", "test4.txt"])

        self.sftp.chdir("sausage")

        self.assertEqual(self.sftp.listdir(), ["test3.txt", "test4.txt"])

        self.sftp.chdir("../..")

        objs  = [ "test1.txt", "potato/test2.txt", "potato/sausage/test3.txt", "potato/sausage/test4.txt", ]
        for obj in objs:
            self.sftp.remove("/%s/%s" % (self.container, obj))

        self.assertEqual(self.sftp.listdir(), [])

    def test_offset_resume(self):
        ''' seek/resume functionality (seek_set) '''
        content_string = "This is a chunk of data"*1024
        self.create_file("testfile.txt", content_string)
        self.assertEquals(self.sftp.stat("testfile.txt").st_size, len(content_string))

        fd = self.sftp.open("testfile.txt", "rb")
        contents = fd.read(1024)
        fd.close()

        fd = self.sftp.open("testfile.txt", "rb")
        fd.seek(1024)
        contents += fd.read(512)
        fd.close()

        fd = self.sftp.open("testfile.txt", "rb")
        fd.seek(1024+512)
        contents += fd.read()
        fd.close()

        self.assertEqual(contents, content_string)
        self.sftp.remove("testfile.txt")

    def tearDown(self):
        self.sftp.close()
        self.transport.close()
        # Delete eveything from the container using the API
        _,fails = self.conn.get_container(self.container)
        for obj in fails:
            self.conn.delete_object(self.container, obj["name"])
        self.conn.delete_container(self.container)
        self.assertEquals(fails, [], "The test failed to clean up after itself leaving these objects: %r" % fails)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""
Main function to setup the daemon process.

Copyright (C) 2011-2013 by Memset Ltd. http://www.memset.com/

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import os
import signal
import sys
import logging
from logging.handlers import SysLogHandler
from ConfigParser import RawConfigParser
from optparse import OptionParser
import daemon
import tempfile
from Crypto import Random
import paramiko
from sftpcloudfs.server import ObjectStorageSFTPServer
from sftpcloudfs.constants import version, project_url, config_file, default_ks_service_type, \
    default_ks_tenant_separator, default_ks_endpoint_type
from ftpcloudfs.fs import ObjectStorageFS

class PIDFile(object):
    """
    PID file implementation using a context manager.

    Entering the context acquires the lock, raising OSError if the file exists
    and it's already locked. Leaving the context cleans the PID file.

    Some methods are implemented for compatibility with lockfile and python-daemon.
    """
    def __init__(self, pidfile=None):
        self.pidfile = pidfile
        self._fd = None

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args, **kwargs):
        self.release()

    def acquire(self):
        pid = os.getpid()
        fd = os.open(self.pidfile, (os.O_CREAT|os.O_EXCL|os.O_WRONLY), 0644)
        self._fd = os.fdopen(fd, "w")
        self._fd.write("%s\n" % pid)
        self._fd.flush()
        return pid

    def is_locked(self):
        try:
            os.stat(self.pidfile)
        except OSError:
            return False
        else:
            return True

    def i_am_locking(self):
        return self._fd != None

    def release(self):
        if self._fd:
            self._fd.close()
            os.remove(self.pidfile)
            self._fd = None

class Main(object):
    def __init__(self):
        """Parse configuration and CLI options."""
        global config_file

        # look for an alternative configuration file
        alt_config_file = False
        # used to show errors before we actually start parsing stuff
        parser = OptionParser()
        for arg in sys.argv:
            if arg == '--config':
                try:
                    alt_config_file = sys.argv[sys.argv.index(arg)+1]
                    config_file = alt_config_file
                except IndexError:
                    pass
            elif arg.startswith('--config='):
                _, alt_config_file = arg.split('=', 1)
                if alt_config_file == '':
                    parser.error("--config option requires an argument")
                config_file = alt_config_file

        config = RawConfigParser({'auth-url': None,
                                  'host-key-file': None,
                                  'bind-address': "127.0.0.1",
                                  'port': 8022,
                                  'memcache': None,
                                  'max-children': "20",
                                  'log-file': None,
                                  'syslog': 'no',
                                  'verbose': 'no',
                                  'scp-support': 'yes',
                                  'pid-file': None,
                                  'uid': None,
                                  'gid': None,
                                  'split-large-files': "0",
                                  # keystone auth 2.0 support
                                  'keystone-auth': False,
                                  'keystone-region-name': None,
                                  'keystone-tenant-separator': default_ks_tenant_separator,
                                  'keystone-service-type': default_ks_service_type,
                                  'keystone-endpoint-type': default_ks_endpoint_type,
                                  })

        if not config.read(config_file) and alt_config_file:
            # the default conf file is optional
            parser.error("failed to read %s" % config_file)

        if not config.has_section('sftpcloudfs'):
            config.add_section('sftpcloudfs')

        parser = OptionParser(version="%prog " + version,
                              description="This is a SFTP interface to OpenStack " + \
                                    "Object Storage (Swift).",
                              epilog="Contact and support at: %s" % project_url)

        parser.add_option("-a", "--auth-url", dest="authurl",
                          default=config.get('sftpcloudfs', 'auth-url'),
                          help="Authentication URL")

        parser.add_option("-k", "--host-key-file", dest="host_key",
                          default=config.get('sftpcloudfs', 'host-key-file'),
                          help="Host RSA key used by the server")

        parser.add_option("-b", "--bind-address", dest="bind_address",
                          default=config.get('sftpcloudfs', 'bind-address'),
                          help="Address to bind (default: 127.0.0.1)")

        parser.add_option("-p", "--port", dest="port",
                          type="int",
                          default=config.get('sftpcloudfs', 'port'),
                          help="Port to bind (default: 8022)")

        memcache = config.get('sftpcloudfs', 'memcache')
        if memcache:
            memcache = [x.strip() for x in memcache.split(',')]
        parser.add_option('--memcache',
                          type="str",
                          dest="memcache",
                          action="append",
                          default=memcache,
                          help="Memcache server(s) to be used for cache (ip:port)")


        parser.add_option("-l", "--log-file", dest="log_file",
                          default=config.get('sftpcloudfs', 'log-file'),
                          help="Log into provided file")

        parser.add_option("-f", "--foreground", dest="foreground",
                          action="store_true",
                          default=False,
                          help="Run in the foreground (don't detach from terminal)")

        parser.add_option("--disable-scp", dest="no_scp",
                          action="store_true",
                          default=not config.getboolean('sftpcloudfs', 'scp-support'),
                          help="Disable SCP support (default: enabled)")

        parser.add_option("--syslog", dest="syslog",
                          action="store_true",
                          default=config.getboolean('sftpcloudfs', 'syslog'),
                          help="Enable logging to system logger (daemon facility)")

        parser.add_option("-v", "--verbose", dest="verbose",
                          action="store_true",
                          default=config.getboolean('sftpcloudfs', 'verbose'),
                          help="Show detailed information on logging")

        parser.add_option('--pid-file',
                          type="str",
                          dest="pid_file",
                          default=config.get('sftpcloudfs', 'pid-file'),
                          help="Pid file location when in daemon mode")

        parser.add_option('--uid',
                          type="int",
                          dest="uid",
                          default=config.get('sftpcloudfs', 'uid'),
                          help="UID to drop the privileges to when in daemon mode")

        parser.add_option('--gid',
                          type="int",
                          dest="gid",
                          default=config.get('sftpcloudfs', 'gid'),
                          help="GID to drop the privileges to when in daemon mode")

        parser.add_option('--keystone-auth',
                          action="store_true",
                          dest="keystone",
                          default=config.get('sftpcloudfs', 'keystone-auth'),
                          help="Use auth 2.0 (Keystone, requires keystoneclient)")

        parser.add_option('--keystone-region-name',
                          type="str",
                          dest="region_name",
                          default=config.get('sftpcloudfs', 'keystone-region-name'),
                          help="Region name to be used in auth 2.0")

        parser.add_option('--keystone-tenant-separator',
                          type="str",
                          dest="tenant_separator",
                          default=config.get('sftpcloudfs', 'keystone-tenant-separator'),
                          help="Character used to separate tenant_name/username in auth 2.0, " + \
                              "default: TENANT%sUSERNAME" % default_ks_tenant_separator)

        parser.add_option('--keystone-service-type',
                          type="str",
                          dest="service_type",
                          default=config.get('sftpcloudfs', 'keystone-service-type'),
                          help="Service type to be used in auth 2.0, default: %s" % default_ks_service_type)

        parser.add_option('--keystone-endpoint-type',
                          type="str",
                          dest="endpoint_type",
                          default=config.get('sftpcloudfs', 'keystone-endpoint-type'),
                          help="Endpoint type to be used in auth 2.0, default: %s" % default_ks_endpoint_type)

        parser.add_option('--config',
                          type="str",
                          dest="config",
                          default=config_file,
                          help="Use an alternative configuration file")

        (options, args) = parser.parse_args()

        # required parameters
        if not options.authurl:
            parser.error("No auth-url provided")

        if not options.host_key:
            parser.error("No host-key-file provided")

        try:
            self.host_key = paramiko.RSAKey(filename=options.host_key)
        except (IOError, paramiko.SSHException), e:
            parser.error("host-key-file: %s" % e)

        if not options.pid_file:
            options.pid_file = "%s/%s.pid" % (tempfile.gettempdir(), __package__)

        if options.memcache:
            ObjectStorageFS.memcache_hosts = options.memcache
            try:
                ObjectStorageFS(None, None, None)
            except (ValueError, TypeError):
                parser.error("memcache: invalid server address, ip:port expected")

        self.pidfile = PIDFile(options.pid_file)
        if self.pidfile.is_locked():
            parser.error("pid-file found: %s\nIs the server already running?" % options.pid_file)

        try:
            options.max_children = int(config.get('sftpcloudfs', 'max-children'))
        except ValueError:
            parser.error('max-children: invalid value, integer expected')

        try:
            options.split_size = int(config.get('sftpcloudfs', 'split-large-files'))*10**6
        except ValueError:
            parser.error('split-large-files: invalid size, integer expected')

        if options.keystone:
            keystone_keys = ('region_name', 'tenant_separator', 'service_type', 'endpoint_type')
            options.keystone = dict((key, getattr(options, key)) for key in keystone_keys)

        self.options = options

    def setup_log(self):
        """Setup server logging facility."""
        self.log = paramiko.util.get_logger("paramiko")

        if self.options.log_file:
            paramiko.util.log_to_file(self.options.log_file)

        if self.options.syslog is True:
            try:
                handler = SysLogHandler(address='/dev/log',
                                        facility=SysLogHandler.LOG_DAEMON)
            except IOError:
                # fall back to UDP
                handler = SysLogHandler(facility=SysLogHandler.LOG_DAEMON)
            finally:
                handler.setFormatter(logging.Formatter('%(name)s[%(_threadid)s]: %(levelname)s: %(message)s'))
                self.log.addHandler(handler)

        if self.options.foreground:
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter('%(asctime)s: %(name)s[%(_threadid)s]: %(levelname)s: %(message)s'))
            self.log.addHandler(handler)

        if self.options.verbose:
            # enable debug for the root logger (used by ObjectStorageFS)
            logging.getLogger().setLevel(logging.DEBUG)
            self.log.setLevel(logging.DEBUG)
            self.log.debug(self.options)
        else:
            self.log.setLevel(logging.INFO)


    def run(self):
        """Run the server."""
        server = ObjectStorageSFTPServer((self.options.bind_address, self.options.port),
                                          host_key=self.host_key,
                                          authurl=self.options.authurl,
                                          max_children=self.options.max_children,
                                          keystone=self.options.keystone,
                                          no_scp=self.options.no_scp,
                                          split_size=self.options.split_size,
                                          )

        dc = daemon.DaemonContext()
        dc.pidfile = self.pidfile

        if self.options.uid:
            dc.uid = self.options.uid

        if self.options.gid:
            dc.gid = self.options.gid

        # FIXME: we don't know the fileno for Random open files, but they're  < 16
        dc.files_preserve = range(server.fileno(), 16)

        if self.options.foreground:
            dc.detach_process = False
            dc.stderr = sys.stderr

        with dc:
            Random.atfork()
            self.setup_log()
            try:
                if os.getuid() == 0:
                    self.log.warning("UID is 0, running as root is not recommended")

                self.log.info("Listening on %s:%s" % (self.options.bind_address, self.options.port))
                server.serve_forever()
            except (SystemExit, KeyboardInterrupt):
                self.log.info("Terminating...")
                if server.active_children:
                    for pid in server.active_children:
                        os.kill(pid, signal.SIGTERM)
                server.server_close()

        if self.pidfile.i_am_locking():
            self.pidfile.release()

        return 0




#!/usr/bin/python
import unittest
import os
import sys
from swiftclient import client
import paramiko

hostname = "127.0.0.1"
port = 8022

# setup logging
#paramiko.util.log_to_file('test_sftpd.log')

class SftpcloudfsTest(unittest.TestCase):
    ''' FTP Cloud FS main test '''

    def setUp(self):
        if not all(['OS_API_KEY' in os.environ,
                    'OS_API_USER' in os.environ,
                    'OS_AUTH_URL' in os.environ,
                    ]):
            print "env OS_API_USER/OS_API_KEY/OS_AUTH_URL not found."
            sys.exit(1)

        self.username = os.environ['OS_API_USER']
        self.api_key = os.environ['OS_API_KEY']
        self.tenant = os.environ.get('OS_API_TENANT')
        self.auth_url = os.environ['OS_AUTH_URL']
        self.container = "sftpcloudfs_testing"

        connection_parameters = {
            "user": self.username,
            "key": self.api_key,
            "authurl": self.auth_url,
        }

        if self.tenant:
            connection_parameters['auth_version'] = 2
            connection_parameters['tenant_name'] = self.tenant

        self.conn = client.Connection(**connection_parameters)
        self.conn.put_container(self.container)

        self.transport = paramiko.Transport((hostname, port))
        self.transport.connect(
            username=self.username,
            password=self.api_key,
            # hostkey=hostkey
        )
        self.channel = self.transport.open_session()
        self.container = "sftpcloudfs_testing"

    def test_setup_and_teardown(self):
        pass

    def test_file_upload(self):
        # Source:
        # https://blogs.oracle.com/janp/entry/how_the_scp_protocol_works
        # Example 1

        self.channel.exec_command('scp -t /%s/foo' % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")


        self.channel.sendall("C0644 6 foo\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")
        self.channel.sendall("Hello\n")

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

        headers, content = self.conn.get_object(self.container, 'foo')
        self.assertEquals(content, 'Hello\n')

    def test_file_upload_to_dir(self):
        self.channel.exec_command('scp -td /%s' % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")


        self.channel.sendall("C0644 6 test\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")
        self.channel.sendall("Hello\n")

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

        headers, content = self.conn.get_object(self.container, 'test')
        self.assertEquals(content, 'Hello\n')

    def test_dir_upload(self):
        # Source:
        # https://blogs.oracle.com/janp/entry/how_the_scp_protocol_works
        # Example 1

        self.channel.exec_command('scp -tr /%s/foodir' % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("D0644 0 foodir\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("C0644 6 test\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("Hello\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("E\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

        headers, content = self.conn.get_object(self.container, 'foodir/test')
        self.assertEquals(content, 'Hello\n')

    def download_file(self):
        self.conn.put_object(self.container, 'bar', 'Hello\n')

        self.channel.exec_command('scp -f /%s/bar' % self.container)
        self.channel.send('\000' * 3)

        response = self.channel.read()

        self.assertEquals(response, 'C0644 6 bar\nHello\n')

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

    def download_dir(self):
        self.conn.put_object(self.container, 'foo/bar', 'Hello\n')

        self.channel.exec_command('scp -fr /%s/foo' % self.container)
        self.channel.send('\000' * 5)

        response = self.channel.read()

        self.assertEquals(
            response,
            'D0644 0 foo\n'
            'C0644 6 bar\nHello\n'
            'E\n'
        )

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

    def test_file_upload_invalid_size(self):
        self.channel.exec_command('scp -t /%s/foo' % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("C0644 not_an_integer test\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\001")

    def test_file_upload_error_disconnect(self):
        self.channel.exec_command('scp -t /%s/foo' % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("C0644 not_an_integer foo\n")
        # shouldn't raise and exception on the server
        # FIXME: test that it doesn't happen!

    def test_file_upload_unicode(self):
        self.channel.exec_command(u'scp -t /%s/Smiley\u263a\ file'.encode("utf-8") % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall(u"C0644 6 Smiley\u263a file\n".encode("utf-8"))
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")
        self.channel.sendall("Hello\n")

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

        headers, content = self.conn.get_object(self.container, u'Smiley\u263a file')
        self.assertEquals(content, 'Hello\n')

    def test_dir_upload_unicode(self):
        self.channel.exec_command(u'scp -tr /%s/Smiley\u263a\ dir'.encode("utf-8") % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall(u"D0644 6 Smiley\u263a dir\n".encode("utf-8"))
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("C0644 6 test\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("Hello\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall("E\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

        headers, content = self.conn.get_object(self.container, u'Smiley\u263a dir/test')
        self.assertEquals(content, 'Hello\n')

    def test_file_upload_like_scp(self):
        self.channel.exec_command('scp -t  -- /%s/file with spaces.txt' % self.container)

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        self.channel.sendall(u"C0644 6 file with spaces.txt\n")
        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")
        self.channel.sendall("Hello\n")

        ack = self.channel.recv(1)
        self.assertEquals(ack, "\000")

        exit_status = self.channel.recv_exit_status()
        self.assertEquals(exit_status, 0)

        tail = self.channel.recv(1)
        self.assertEquals(tail, '')

        headers, content = self.conn.get_object(self.container, 'file with spaces.txt')
        self.assertEquals(content, 'Hello\n')

    def tearDown(self):
        self.channel.close()
        self.transport.close()

        # Delete eveything from the container using the API
        _, fails = self.conn.get_container(self.container)
        for obj in fails:
            self.conn.delete_object(self.container, obj["name"])

        self.conn.delete_container(self.container)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from setuptools import setup, find_packages
from sftpcloudfs.constants import version, project_url

def readme():
    try:
        return open('README.md').read()
    except:
        return ""

setup(name='sftp-cloudfs',
      version=version,
      description='SFTP interface to OpenStack Object Storage (Swift)',
      long_description=readme(),
      author='Nick Craig-Wood',
      author_email='nick@memset.com',
      url=project_url,
      license='MIT',
      include_package_data=True,
      zip_safe=False,
      install_requires=['paramiko>=1.7.6', 'python-swiftclient>=1.6', 'python-daemon>=1.5',
                        'python-memcached>=1.45', 'ftp-cloudfs>=0.25.2'],
      scripts=['bin/sftpcloudfs'],
      packages = find_packages(exclude=['tests']),
      tests_require = ["nose"],
      classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Environment :: No Input/Output (Daemon)',
        'License :: OSI Approved :: MIT License',
        ],
      test_suite = "nose.collector",
      )



version = "0.12.2"
project_url = "https://pypi.python.org/pypi/sftp-cloudfs/"
config_file = "/etc/sftpcloudfs.conf"

# keystone defaults
default_ks_tenant_separator = '.'
default_ks_service_type = 'object-store'
default_ks_endpoint_type = 'publicURL'



import stat
import optparse
import posixpath
import socket
import threading

from ftpcloudfs.fs import IOSError, parse_fspath

class SCPException(Exception):
    def __init__(self, status, message):
        self.status = status
        super(Exception, self).__init__(message)


class SCPHandler(threading.Thread):

    CHUNK_SIZE = 64*1024
    TIMEOUT = 30.0 # seconds

    def __init__(self, arguments, channel, fs, log):
        super(SCPHandler, self).__init__()
        self.log = log
        self.channel = channel
        self.channel.settimeout(self.TIMEOUT)
        self.fs = fs
        self.args = arguments
        self.buffer = ""

    @classmethod
    def get_argparser(cls):
        parser = optparse.OptionParser(
            prog='scp',
            description='Remote part of secure copy tool'
        )
        parser.add_option('-v', action='count', dest='verbose',
                          help='makes SCP verbose')
        parser.add_option('-t', action='store_true', dest='copy_to')
        parser.add_option('-f', action='store_true', dest='copy_from')
        parser.add_option('-r', action='store_true', dest='recursive',
                          help='Recursively copy entire directories.')
        # unused
        parser.add_option('-p', action='store_true', dest='preserve',
                          help='Preserves modification times, access '
                               'times, and modes from the original file.')
        parser.add_option('-d', action='store_true', dest='directory',
                          help='Target should be a directory')
        parser.add_option('-E', action='store_true', dest='xargs',
                          help='Target should be a directory')

        def ap_exit(status=0, message=""):
            raise SCPException(status, message)

        parser.exit = ap_exit
        parser.error = lambda msg: ap_exit(2, msg)

        return parser

    def run(self):
        try:
            self.args, self.paths = SCPHandler.get_argparser().parse_args(self.args)
            self.log.debug("SCP %r", self.args)

            if self.args.copy_to and self.args.copy_from:
                raise SCPException(4, "-t and -f can't be combined")

            if len(self.paths) != 1:
                raise SCPException(4, "scp takes exactly one path")

            if self.args.copy_to:
                self.receive()
            elif self.args.copy_from:
                path = self.paths[0]
                try:
                    path_stat = self.fs.stat(path)
                except IOSError, ex:
                    raise SCPException(1, ex)

                self.send(path, path_stat)
            else:
                raise SCPException(4, "Missing -t or -f argument")
        except SCPException, ex:
            self.log.info("SCP reject: %s", ex)
            self.send_status_and_close(msg=ex, status=ex.status)
        except socket.timeout:
            self.log.info("SCP timeout")
            self.send_status_and_close(msg="%ss timeout" % self.TIMEOUT, status=1)
        except:
            self.log.exception("SCP internal exception")
            self.send_status_and_close(msg="internal error", status=1)
        else:
            self.send_status_and_close()

    def send_status_and_close(self, msg=None, status=0):
        try:
            if msg:
                self.channel.sendall('\x01scp: ')
                self.channel.sendall(str(msg))
                self.channel.sendall('\n')
            self.channel.send_exit_status(status)
        except socket.error, ex:
            self.log.warn("Failed to properly close the channel: %r" % ex.message)
        finally:
            try:
                self.channel.close()
            except socket.error:
                pass

    def recv(self, size):
        if self.buffer:
            result = self.buffer[:size]
            self.buffer = self.buffer[size:]
            return result

        return self.channel.recv(size)

    def recv_line(self):
        if '\n' not in self.buffer:
            while True:
                chunk = self.channel.recv(1024)
                self.buffer += chunk
                if '\n' in chunk:
                    break

        line, self.buffer = self.buffer.split('\n', 1)
        return line

    def receive(self):
        # Ack the connection
        self.channel.send('\x00')

        if self.args.directory:
            directory = self.paths[0]
            filename = None
        else:
            directory, filename = posixpath.split(self.paths[0])

        if not self.fs.isdir(directory):
            raise SCPException(1, '%s is not a directory' % directory)

        record = self.recv_line()
        self.receive_inner(directory, record, override_name=filename)

    def receive_inner(self, path, record, override_name=None):

        if record[0] == 'T':
            # We're going to ignore this record, but for compatibility, we'll
            # confirm its arrival.

            # mtime, mtime_u, atime, atime_u = record[1:].split()
            # mtime = float(mtime) + float(mtime_u) / 100000
            # atime = float(atime) + float(atime_u) / 100000

            # ACK this file record
            self.channel.send('\x00')
            record = self.recv_line()

        elif record[0] == 'C':
            mode, size, name = record[1:].split(' ', 2)
            try:
                size = int(size)
            except ValueError:
                raise SCPException(1, 'invalid size')
            target_path = path.rstrip('/') + '/' + (override_name or name)

            if self.fs.isdir(target_path):
                raise SCPException(1, '%s: directory exists' % target_path)

            # we can't create files on the root, only inside a container
            if not all(parse_fspath(target_path)):
                raise SCPException(1, "%s: container required" % target_path)

            # ACK this file record
            self.channel.send('\x00')

            fd = self.fs.open(target_path, 'w')

            bytes_sent = 0
            while bytes_sent < size:
                chunk = self.recv(self.CHUNK_SIZE)
                fd.write(chunk)
                bytes_sent += len(chunk)

            fd.close()
            # ACK sending this file
            self.channel.send('\x00')
            #self.wait_for_ack()

        elif record[0] == 'D':
            mode, size, name = record[1:].split(' ', 2)

            target_path = path.rstrip('/') + '/' + (override_name or name)

            if self.fs.isfile(target_path):
                raise SCPException(1, '%s: file exists', target_path)

            # ACK this directory record
            self.channel.send('\x00')

            self.fs.mkdir(target_path)

            while True:
                record = self.recv_line()
                if record[0] == 'E':
                    # ACK this file record
                    self.channel.send('\x00')
                    break
                else:
                    self.receive_inner(target_path, record)

    def send(self, path, path_stat):
        self.log.debug('About to send %s', path)

        if self.args.preserve:
            self.channel.sendall("T%i 0 %i 0\n" % (
                                 path_stat.st_mtime,
                                 path_stat.st_atime))
            self.wait_for_ack()

        if stat.S_ISREG(path_stat.st_mode):
            self.channel.sendall("C%04o %i %s\n" % (
                                 path_stat.st_mode & 07777,
                                 path_stat.st_size,
                                 posixpath.basename(path)))
            self.wait_for_ack()

            fd = self.fs.open(path, 'r')
            while True:
                chunk = fd.read(self.CHUNK_SIZE)
                if chunk:
                    self.channel.sendall(chunk)
                else:
                    break

            # signal the end of the transfer
            self.channel.send('\x00')
            self.wait_for_ack()

        elif not self.args.recursive:
            self.channel.sendall("scp: %s is not a regular file\n" % path)
            return 1
        else:
            self.channel.send("D%04o %i %s\n" % (
                              path_stat.st_mode & 07777,
                              0,
                              posixpath.basename(path)))

            self.wait_for_ack()

            for subpath, subpath_stat in self.fs.listdir_with_stat(path):
                subpath = path + "/" + subpath
                self.send_inner(subpath, subpath_stat)

            self.channel.send("E\n")
            self.wait_for_ack()

    def wait_for_ack(self):
        """ Wait for the ack byte """
        ack = self.channel.recv(1)
        if ack != '\x00':
            raise Exception("Command not acknowledged (%r)" % ack)

